@@ -46,19 +46,22 @@ extern"C" {
 #endif
 #include <stdbool.h>
 #include <stdatomic.h>
+#include <semaphore.h>
 #include <zlib.h>
 #include <lz4frame.h>
+#include <arpa/inet.h>
+#include <limits.h>
+#include "qz_utils.h"
 
 /**
  *  define release version
  */
-#define QATZIP_VERSION "1.2.0"
+#define QATZIP_VERSION "1.3.1"
 
 #define SUCCESS              1
 #define FAILURE              0
 #define QZ_WAIT_SW_PENDING   10
 
-#define NODE_0               0
 #define NUM_BUFF             (32)
 /**
  * For less than 8K hardware buffer size, it needs more in-flight buffers
@@ -129,11 +132,25 @@ extern"C" {
 #define QZ_LZ4_STOREDBLOCK_FLAG 0x80000000U
 #define QZ_LZ4_STORED_HEADER_SIZE 4
 
+//https://stackoverflow.com/questions/9050260/what-does-a-zlib-header-look-like
+//https://datatracker.ietf.org/doc/html/rfc1950
+//zlib header CMF
+#define QZ_ZLIB_HEADER_CMF              0x78
+// zlib magic header FLG with different compression level
+#define QZ_ZLIB_HEADER_FLG_LOW          0x01
+#define QZ_ZLIB_HEADER_FLG_FAST         0x5E
+#define QZ_ZLIB_HEADER_FLG_DEFAULT      0x9C
+#define QZ_ZLIB_HEADER_FLG_BEST         0xDA
+
+
 #define DATA_FORMAT_DEFAULT     DEFLATE_GZIP_EXT
 #define IS_DEFLATE_RAW(fmt)  (DEFLATE_RAW == (fmt))
 #define IS_DEFLATE(fmt) \
         (DEFLATE_RAW == (fmt) || DEFLATE_GZIP == (fmt) || \
-        DEFLATE_GZIP_EXT == (fmt) || DEFLATE_4B == (fmt))
+        DEFLATE_GZIP_EXT == (fmt) || DEFLATE_4B == (fmt) || DEFLATE_ZLIB == (fmt))
+
+struct QzAsyncReq_S;
+typedef struct QzAsyncReq_S QzAsyncReq_T;
 
 typedef struct QzCpaStream_S {
     signed long seq;
@@ -150,6 +167,7 @@ typedef struct QzCpaStream_S {
     unsigned int checksum;
     unsigned int orgdatalen;
     CpaDcOpData opData;
+    QzAsyncReq_T *req;
 } QzCpaStream_T;
 
 typedef struct QzInstance_S {
@@ -198,6 +216,8 @@ typedef struct ProccesData_S {
     atomic_char qat_available;
     /* this thread handler is for keep polling device fatal events.*/
     pthread_t t_poll_heartbeat;
+    /* Define pthread key here, for different thread local variable */
+    pthread_key_t async_req_key;
 } processData_T;
 
 typedef enum {
@@ -214,6 +234,7 @@ typedef enum {
 } DeflateState_T;
 
 // Include all support data format
+// Don't change the enumeration order
 typedef enum DataFormatInternal_E {
     DEFLATE_4B = 0,
     /**< Data is in raw deflate format with 4 byte header */
@@ -227,6 +248,8 @@ typedef enum DataFormatInternal_E {
     /**< Data is in LZ4 format with frame headers */
     LZ4S_BK,
     /**< Data is in LZ4s sequences with block header, it's only for post processed */
+    DEFLATE_ZLIB,
+    /**< Data is in  deflate wrapped by zlib header and footer */
 } DataFormatInternal_T;
 
 // Include all support session parameters
@@ -275,8 +298,64 @@ typedef struct QzSessionParamsInternal_S {
     /**< 0 means disable sensitive mode, 1 means enable sensitive mode*/
     unsigned int lz4s_mini_match;
     /**< Set lz4s dictionary mini match, which would be 3 or 4 */
+    unsigned char stop_decompression_stream_end;
+    /* stop decompression at end the stream, default is 0 */
+    unsigned char zlib_format;
 } QzSessionParamsInternal_T;
 
+/* lsm_met_len_shift is global variable */
+#define LSM_MET_DEPTH (1<<(lsm_met_len_shift))
+
+typedef struct LatencyMetrix_S {
+    unsigned long *latency_array;
+    unsigned long arr_idx;
+    unsigned long arr_total;
+    unsigned long arr_avg;
+    unsigned long invoke_counter;
+#ifdef QATZIP_DEBUG
+    unsigned long sess_lat_total;
+    unsigned long sess_lat_avg;
+    unsigned long max_latency;
+    unsigned long min_latency;
+    unsigned long max_cnt;
+    unsigned long min_cnt;
+#endif
+} LatencyMetrix_T;
+
+/* This mask should change according to QzAsyncOperationType_E enum */
+#define ASYNC_POLLING_MASK 1
+#define DEFAULT_ASYNC_QUEUE_SIZE 1024
+
+/* compress is even, decompress is odd */
+typedef enum QzAsyncOperationType_E {
+    QZ_COMPRESS = 0,
+    QZ_DECOMPRESS = 1,
+} QzAsyncOperationType_T;
+
+struct QzAsyncReq_S {
+    QzSession_T *sess;
+    const unsigned char *src;
+    unsigned char *dest;
+    QzResult_T *qzResults;
+    qzAsyncCallbackFn qzAsyncallback;
+    QzAsyncOperationType_T op_type;
+
+    const unsigned char *req_src;
+    unsigned int req_in_len;
+    unsigned char *req_dest;
+    unsigned int req_out_len;
+};
+
+typedef struct QzAsynctrl_S {
+    int async_ctrl_init;
+    QzRing_T *async_req_ring;
+    pthread_key_t async_req_key;
+    pthread_t async_consume_t;
+    pthread_t async_polling_t;
+    int async_polling_direct;
+    sem_t sem;
+} QzAsynctrl_T;
+
 typedef struct QzSess_S {
     int inst_hint;   /*which instance we last used*/
     QzSessionParamsInternal_T sess_params;
@@ -314,6 +393,15 @@ typedef struct QzSess_S {
     z_stream *deflate_strm;
     DeflateState_T deflate_stat;
     LZ4F_dctx *dctx;
+    void *qzdeflateExtData;
+    /**< An opaque pointer containing extended results for deflate session*/
+
+    /* LSM */
+    LatencyMetrix_T RRT;
+    LatencyMetrix_T PPT;
+    LatencyMetrix_T SWT;
+    /* Async mode */
+    QzAsynctrl_T *async_ctrl;
 } QzSess_T;
 
 typedef struct QzStreamBuf_S {
@@ -375,6 +463,21 @@ typedef struct QzMem_S {
     int numa;
 } QzMem_T;
 
+// custom deflateExt data structure
+typedef struct QzDeflateExtCustomData_S {
+    unsigned char end_of_stream;
+} QzDeflateExtCustomData_T;
+
+typedef struct StdZLIBH_S {
+    // (CMF) defines compression method and flags (in case of "deflate", the window size).
+    unsigned char cmf;
+    unsigned char flags;
+} StdZlibH_T;
+
+typedef struct StdZlibF_T {
+    uint32_t adler32; /* adler32 checksum */
+} StdZlibF_T;
+
 #pragma pack(push, 1)
 /* lz4 frame header */
 typedef struct QzLZ4H_S {
@@ -399,6 +502,8 @@ unsigned long qzGzipHeaderSz(void);
 unsigned long qz4BHeaderSz(void);
 unsigned long stdGzipHeaderSz(void);
 unsigned long stdGzipFooterSz(void);
+unsigned long stdZlibHeaderSz(void);
+unsigned long stdZlibFooterSz(void);
 unsigned long outputHeaderSz(DataFormatInternal_T data_fmt);
 unsigned long outputFooterSz(DataFormatInternal_T data_fmt);
 void qzGzipHeaderGen(unsigned char *ptr, CpaDcRqResults *res);
@@ -409,10 +514,14 @@ void outputHeaderGen(unsigned char *ptr,
                      CpaDcRqResults *res,
                      DataFormatInternal_T data_fmt);
 void qzGzipFooterGen(unsigned char *ptr, CpaDcRqResults *res);
-void outputFooterGen(QzSess_T *qz_sess,
+void outputFooterGen(unsigned char *ptr,
                      CpaDcRqResults *res,
                      DataFormatInternal_T data_fmt);
 void qzGzipFooterExt(const unsigned char *const ptr, StdGzF_T *ftr);
+void stdZlibHeaderGen(unsigned char *ptr, CpaDcRqResults *res);
+int qzVerifyZlibHeader(const unsigned char *ptr,
+                       const unsigned int *const src_len);
+void qzZlibFooterGen(unsigned char *ptr, CpaDcRqResults *res);
 
 int isQATProcessable(const unsigned char *ptr,
                      const unsigned int *const src_len,
@@ -431,6 +540,8 @@ int qzSWDecompressMulti(QzSession_T *sess, const unsigned char *src,
                         unsigned int *compressed_buffer_len);
 
 unsigned char getSwBackup(QzSession_T *sess);
+void setDeflateEndOfStream(QzSess_T *sess, unsigned char val);
+unsigned char getDeflateEndOfStream(QzSess_T *qz_sess);
 
 #ifdef ADF_PCI_API
 extern CpaStatus icp_adf_get_numDevices(Cpa32U *);
@@ -442,14 +553,13 @@ void *qzMemSet(void *ptr, unsigned char filler, unsigned int count);
 
 unsigned char *findStdGzipFooter(const unsigned char *src_ptr,
                                  long src_avail_len);
-
 void removeSession(int i);
 
 void cleanUpInstMem(int i);
 
 void qzMemDestory(void);
 
-void streamBufferCleanup();
+void streamBufferCleanup(void);
 
 //lz4 functions
 unsigned long qzLZ4HeaderSz(void);
@@ -475,6 +585,7 @@ int qzCheckParams(QzSessionParams_T *params);
 int qzCheckParamsDeflate(QzSessionParamsDeflate_T *params);
 int qzCheckParamsLZ4(QzSessionParamsLZ4_T *params);
 int qzCheckParamsLZ4S(QzSessionParamsLZ4S_T *params);
+int qzCheckParamsDeflateExt(QzSessionParamsDeflateExt_T *params);
 
 void qzGetParams(QzSessionParamsInternal_T *internal_params,
                  QzSessionParams_T *params);
@@ -484,6 +595,8 @@ void qzGetParamsLZ4(QzSessionParamsInternal_T *internal_params,
                     QzSessionParamsLZ4_T *params);
 void qzGetParamsLZ4S(QzSessionParamsInternal_T *internal_params,
                      QzSessionParamsLZ4S_T *params);
+void qzGetParamsDeflateExt(QzSessionParamsInternal_T *internal_params,
+                           QzSessionParamsDeflateExt_T *params);
 
 void qzSetParamsLZ4S(QzSessionParamsLZ4S_T *params,
                      QzSessionParamsInternal_T *internal_params);
@@ -493,7 +606,8 @@ void qzSetParamsDeflate(QzSessionParamsDeflate_T *params,
                         QzSessionParamsInternal_T *internal_params);
 void qzSetParams(QzSessionParams_T *params,
                  QzSessionParamsInternal_T *internal_params);
-
+void qzSetParamsDeflateExt(QzSessionParamsDeflateExt_T *params,
+                           QzSessionParamsInternal_T *internal_params);
 /*  SW Fallback for HW request offload failed or
     HW respond in Error status
  */
@@ -567,5 +681,43 @@ void decompOutSkipErrorRespond(int i, int j, QzSess_T *qz_sess);
 int decompOutCheckSum(int i, int j, QzSession_T *sess,
                       CpaDcRqResults *resl);
 
+/* LSM */
+static inline uint64_t rdtsc()
+{
+    unsigned int lo, hi;
+    __asm__ __volatile__("rdtsc" : "=a"(lo), "=d"(hi));
+    return ((uint64_t)hi << 32) | lo;
+}
 
+void metrixReset(LatencyMetrix_T *m);
+void metrixUpdate(LatencyMetrix_T *m, unsigned long val);
+int compLSMFallback(QzSession_T *sess, const unsigned char *src,
+                    unsigned int *src_len, unsigned char *dest,
+                    unsigned int *dest_len, unsigned int last);
+int decompLSMFallback(QzSession_T *sess, const unsigned char *src,
+                      unsigned int *src_len, unsigned char *dest,
+                      unsigned int *dest_len);
+
+/* Internal Async API */
+int AsyncCompOutCheckDestLen(int i, int j, QzSession_T *sess,
+                             long dest_receive_sz);
+void AsyncCompOutValidDestBufferCleanUp(int i, int j,
+                                        unsigned int dest_receive_sz);
+int AsyncDecompOutSWFallback(int i, int j, QzSession_T *sess,
+                             QzAsyncReq_T *req);
+
+void AsyncDecompOutValidDestBufferCleanUp(int i, int j, QzSess_T *qz_sess,
+        CpaDcRqResults *resl,
+        QzAsyncReq_T *req);
+
+int populateAsyncReq(QzSession_T *sess, const unsigned char *src,
+                     unsigned char *dest, qzAsyncCallbackFn callback,
+                     QzResult_T *qzResults, QzAsyncOperationType_T op_type,
+                     QzAsyncReq_T *req);
+
+int qzAsyncReqSubmit(QzSession_T *sess, QzAsyncReq_T *req,
+                     QzDirection_T direct);
+int qzAsyncSetupHWSession(QzSession_T *sess, QzDirection_T direct);
+int qzSetupAsyncCtrl(QzSession_T *sess);
+int GetStableInstance(QzSession_T *sess);
 #endif //_QATZIPP_H
