@@ -133,6 +133,9 @@ int qzDeflateSWCompress(QzSession_T *sess, const unsigned char *src,
         case DEFLATE_RAW:
             windows_bits = -MAX_WBITS;
             break;
+        case DEFLATE_ZLIB:
+            windows_bits = MAX_WBITS;
+            break;
         case DEFLATE_GZIP:
         case DEFLATE_GZIP_EXT:
         default:
@@ -294,12 +297,15 @@ int qzDeflateSWDecompress(QzSession_T *sess, const unsigned char *src,
     *src_len = 0;
     *dest_len = 0;
 
-    QZ_DEBUG("decomp_sw data_fmt: %d\n", data_fmt);
+    QZ_INFO("decomp_sw data_fmt: %d\n", data_fmt);
     switch (data_fmt) {
     case DEFLATE_4B:
     case DEFLATE_RAW:
         windows_bits = -MAX_WBITS;
         break;
+    case DEFLATE_ZLIB:
+        windows_bits = MAX_WBITS;
+        break;
     case DEFLATE_GZIP:
     case DEFLATE_GZIP_EXT:
     default:
@@ -319,7 +325,7 @@ int qzDeflateSWDecompress(QzSession_T *sess, const unsigned char *src,
             ret = QZ_FAIL;
             goto done;
         }
-        QZ_DEBUG("\n****** inflate init done with win_bits: %d *****\n", windows_bits);
+        QZ_INFO("\n****** inflate init done with win_bits: %d *****\n", windows_bits);
         qz_sess->inflate_stat = InflateInited;
         stream->total_in = 0;
         total_in = 0;
@@ -331,10 +337,11 @@ int qzDeflateSWDecompress(QzSession_T *sess, const unsigned char *src,
     }
 
     zlib_ret = inflate(stream, Z_SYNC_FLUSH);
+    setDeflateEndOfStream(qz_sess, 0);
     switch (zlib_ret) {
     case Z_OK:
         if (QZ_LOW_DEST_MEM == sess->thd_sess_stat) {
-            QZ_DEBUG("ERR: inflate failed with Z_DATA_ERROR\n");
+            QZ_INFO("ERR: inflate failed with Z_DATA_ERROR\n");
             ret = QZ_DATA_ERROR;
             qz_sess->inflate_stat = InflateError;
             goto done;
@@ -345,14 +352,15 @@ int qzDeflateSWDecompress(QzSession_T *sess, const unsigned char *src,
     case Z_STREAM_END:
         ret = QZ_OK;
         qz_sess->inflate_stat = InflateEnd;
+        setDeflateEndOfStream(qz_sess, 1);
         break;
     case Z_DATA_ERROR:
-        QZ_DEBUG("ERR: inflate failed with Z_DATA_ERROR\n");
+        QZ_INFO("ERR: inflate failed with Z_DATA_ERROR\n");
         ret = QZ_DATA_ERROR;
         qz_sess->inflate_stat = InflateError;
         goto done;
     default:
-        QZ_DEBUG("ERR: inflate failed with error code %d\n", ret);
+        QZ_INFO("ERR: inflate failed with error code %d\n", ret);
         ret = QZ_FAIL;
         qz_sess->inflate_stat = InflateError;
         goto done;
@@ -373,11 +381,11 @@ done:
              *dest_len);
     if (zlib_ret == Z_STREAM_END || QZ_LOW_DEST_MEM == sess->thd_sess_stat) {
         if (Z_OK != inflateEnd(stream)) {
-            QZ_DEBUG("inflateEnd failed.\n");
+            QZ_INFO("inflateEnd failed.\n");
             ret = QZ_FAIL;
         }
         qz_sess->inflate_stat = InflateNull;
-        QZ_DEBUG("\n****** inflate end done *****\n");
+        QZ_INFO("\n****** inflate end done *****\n");
     }
 
     return ret;
@@ -394,11 +402,12 @@ int qzSWDecompressMultiGzip(QzSession_T *sess, const unsigned char *src,
     const unsigned int output_len = *dest_len;
     unsigned int cur_input_len = input_len;
     unsigned int cur_output_len = output_len;
+    QzSess_T *qz_sess = (QzSess_T *) sess->internal;
 #ifdef QATZIP_DEBUG
     insertThread((unsigned int)pthread_self(), DECOMPRESSION, SW);
 #endif
-    QZ_DEBUG("Start qzSWDecompressMultiGzip: src_len %u dest_len %u\n",
-             *src_len, *dest_len);
+    QZ_INFO("Start qzSWDecompressMultiGzip: src_len %u dest_len %u\n",
+            *src_len, *dest_len);
 
     *src_len = 0;
     *dest_len = 0;
@@ -419,11 +428,15 @@ int qzSWDecompressMultiGzip(QzSession_T *sess, const unsigned char *src,
         cur_output_len = output_len - total_out;
         *src_len  = total_in;
         *dest_len = total_out;
+        if (getDeflateEndOfStream(qz_sess) == 1 &&
+            qz_sess->sess_params.stop_decompression_stream_end == 1) {
+            goto out;
+        }
     }
 
 out:
-    QZ_DEBUG("Exit qzSWDecompressMultiGzip: src_len %u dest_len %u\n",
-             *src_len, *dest_len);
+    QZ_INFO("Exit qzSWDecompressMultiGzip: src_len %u dest_len %u\n",
+            *src_len, *dest_len);
     return ret;
 }
 
@@ -446,8 +459,8 @@ int qzLZ4SWCompress(QzSession_T *sess, const unsigned char *src,
         goto lz4_compress_fail;
     }
     *dest_len = total_out;
-    QZ_DEBUG("Exit qzLZ4SWCompress: src_len %u dest_len %u\n",
-             *src_len, *dest_len);
+    QZ_INFO("Exit qzLZ4SWCompress: src_len %u dest_len %u\n",
+            *src_len, *dest_len);
 
     return QZ_OK;
 
@@ -465,8 +478,8 @@ int qzLZ4SWDecompress(QzSession_T *sess, const unsigned char *src,
     size_t out_sz = 0;
     size_t ret = QZ_OK;
     QzSess_T *qz_sess = NULL;
-    QZ_DEBUG("Enter qzLZ4SWDecompress: src_len %u dest_len %u\n",
-             *src_len, *dest_len);
+    QZ_INFO("Enter qzLZ4SWDecompress: src_len %u dest_len %u\n",
+            *src_len, *dest_len);
 
     qz_sess = (QzSess_T *) sess->internal;
     if (qz_sess->dctx == NULL) {
@@ -498,13 +511,13 @@ int qzLZ4SWDecompress(QzSession_T *sess, const unsigned char *src,
          * it needs more compressed data to decompress. The remaining compressed data
          * was stored in dctx.
          */
-        QZ_DEBUG("LZ4F_decompress: incomplete compressed data, need more data\n");
+        QZ_INFO("LZ4F_decompress: incomplete compressed data, need more data\n");
     }
 
     *src_len = in_sz;
     *dest_len = out_sz;
-    QZ_DEBUG("Exit qzLZ4SWDecompress: src_len %u dest_len %u\n",
-             *src_len, *dest_len);
+    QZ_INFO("Exit qzLZ4SWDecompress: src_len %u dest_len %u\n",
+            *src_len, *dest_len);
 
     return QZ_OK;
 
@@ -533,8 +546,8 @@ int qzSWDecompressMultiLZ4(QzSession_T *sess, const unsigned char *src,
 #ifdef QATZIP_DEBUG
     insertThread((unsigned int)pthread_self(), DECOMPRESSION, SW);
 #endif
-    QZ_DEBUG("Start qzSWDecompressMultiLz4: src_len %u dest_len %u\n",
-             *src_len, *dest_len);
+    QZ_INFO("Start qzSWDecompressMultiLz4: src_len %u dest_len %u\n",
+            *src_len, *dest_len);
 
     *src_len = 0;
     *dest_len = 0;
@@ -558,8 +571,8 @@ int qzSWDecompressMultiLZ4(QzSession_T *sess, const unsigned char *src,
     }
 
 out:
-    QZ_DEBUG("Exit qzSWDecompressMultiLz4: src_len %u dest_len %u\n",
-             *src_len, *dest_len);
+    QZ_INFO("Exit qzSWDecompressMultiLz4: src_len %u dest_len %u\n",
+            *src_len, *dest_len);
     return ret;
 }
 
@@ -587,6 +600,7 @@ int qzSWCompress(QzSession_T *sess, const unsigned char *src,
     case DEFLATE_GZIP:
     case DEFLATE_GZIP_EXT:
     case DEFLATE_4B:
+    case DEFLATE_ZLIB:
         ret = qzDeflateSWCompress(sess, src, src_len, dest, dest_len, last);
         break;
     case LZ4_FH:
@@ -625,6 +639,7 @@ int qzSWDecompress(QzSession_T *sess, const unsigned char *src,
     case DEFLATE_GZIP:
     case DEFLATE_GZIP_EXT:
     case DEFLATE_4B:
+    case DEFLATE_ZLIB:
         ret = qzDeflateSWDecompress(sess, src, src_len, dest, dest_len);
         break;
     case LZ4_FH:
@@ -663,6 +678,7 @@ int qzSWDecompressMulti(QzSession_T *sess, const unsigned char *src,
     case DEFLATE_GZIP:
     case DEFLATE_GZIP_EXT:
     case DEFLATE_4B:
+    case DEFLATE_ZLIB:
         ret = qzSWDecompressMultiGzip(sess, src, src_len, dest, dest_len);
         break;
     case LZ4_FH:
@@ -686,19 +702,19 @@ int compInSWFallback(int i, int j, QzSession_T *sess,
     QzSess_T *qz_sess = (QzSess_T *)sess->internal;
 
     if (!qz_sess->sess_params.sw_backup) {
-        QZ_DEBUG("The instance %d heartbeat is failure, Don't enable sw fallback, compressIn fatal ERROR!\n",
+        QZ_ERROR("The instance %d heartbeat down, Don't enable sw fallback, compressIn error!\n",
                  i);
         return QZ_FAIL;
     }
 
     if (qz_sess->single_thread) {
-        QZ_DEBUG("The instance %d failure, single_thread, compressIn fatal ERROR!\n",
-                 i);
+        QZ_INFO("The instance %d fallback to sw, single_thread, back to API level fallback!\n",
+                i);
         return QZ_FAIL;
     }
 
     if (qz_sess->stop_submitting) {
-        QZ_DEBUG("compInSWFallback stop submit\n");
+        QZ_INFO("compInSWFallback stop submit\n");
         return QZ_FAIL;
     }
 
@@ -717,7 +733,7 @@ int compInSWFallback(int i, int j, QzSession_T *sess,
              src_ptr, qz_sess->next_dest, src_send_sz, qz_sess->seq);
 
     if (unlikely(QZ_OK != rc)) {
-        QZ_ERROR("SW CompIn fallback failure! compress fatal ERROR!\n");
+        QZ_ERROR("SW CompIn fallback failure! compress error!\n");
         return QZ_FAIL;
     }
 
@@ -744,7 +760,7 @@ int compOutSWFallback(int i, int j, QzSession_T *sess,
     unsigned char *dest_ptr = g_process.qz_inst[i].dest_buffers[j]->pBuffers->pData;
 
     if (!qz_sess->sess_params.sw_backup) {
-        QZ_DEBUG("The instance %d heartbeat is failure, Don't enable sw fallback, compressOut fatal ERROR!\n",
+        QZ_ERROR("The instance %d heartbeat down, Don't enable sw fallback, compressOut error!\n",
                  i);
         return QZ_FAIL;
     }
@@ -756,7 +772,7 @@ int compOutSWFallback(int i, int j, QzSession_T *sess,
     rc = qzSWCompress(sess, src_ptr, &src_send_sz, dest_ptr, &dest_receive_sz,
                       qz_sess->last);
     if (QZ_OK != rc) {
-        QZ_ERROR("SW CompOut fallback failure! compress fatal ERROR!\n");
+        QZ_ERROR("SW CompOut fallback failure! compress error!\n");
         return QZ_FAIL;
     }
 
@@ -783,21 +799,21 @@ int decompInSWFallback(int i, int j, QzSession_T *sess,
     QzSess_T *qz_sess = (QzSess_T *)sess->internal;
 
     if (!qz_sess->sess_params.sw_backup) {
-        QZ_DEBUG("The instance %d heartbeat is failure, Don't enable sw fallback, decompressIn fatal ERROR!\n",
+        QZ_ERROR("The instance %d heartbeat down, Don't enable sw fallback, decompressIn error!\n",
                  i);
         sess->thd_sess_stat = QZ_FAIL;
         return QZ_FAIL;
     }
 
     if (qz_sess->single_thread) {
-        QZ_DEBUG("The instance %d failure, single thread, decompressIn fatal ERROR!\n",
-                 i);
+        QZ_INFO("The instance %d fallback to sw, single_thread, back to API level fallback!\n",
+                i);
         sess->thd_sess_stat = QZ_FAIL;
         return QZ_FAIL;
     }
 
     if (qz_sess->stop_submitting) {
-        QZ_DEBUG("decompInSWFallback stop submit\n");
+        QZ_INFO("decompInSWFallback stop submit\n");
         return QZ_FAIL;
     }
 
@@ -806,14 +822,14 @@ int decompInSWFallback(int i, int j, QzSession_T *sess,
     }
 
     QZ_DEBUG("SW DecompIn src_ptr %p, dest_ptr %p, Sending %u bytes, seq = %ld, instance = %d\n",
-             src_ptr, dest_ptr, tmp_src_avail_len, qz_sess->seq, i);
+             src_ptr, dest_ptr, *tmp_src_avail_len, qz_sess->seq, i);
     rc = qzSWDecompress(sess,
                         src_ptr,
                         tmp_src_avail_len,
                         dest_ptr,
                         tmp_dest_avail_len);
     if (unlikely(QZ_OK != rc)) {
-        QZ_ERROR("SW deCompIn fallback failure! decompress fatal ERROR!\n");
+        QZ_ERROR("SW deCompIn fallback failure! decompress error!\n");
         sess->thd_sess_stat = QZ_FAIL;
         return QZ_FAIL;
     }
@@ -844,7 +860,7 @@ int decompOutSWFallback(int i, int j, QzSession_T *sess,
     src_send_sz += (outputHeaderSz(data_fmt) + outputFooterSz(data_fmt)) ;
 
     if (!qz_sess->sess_params.sw_backup) {
-        QZ_DEBUG("The instance %d heartbeat is failure, Don't enable sw fallback, decompressOut fatal ERROR!\n",
+        QZ_ERROR("The instance %d heartbeat down, Don't enable sw fallback, decompressOut error!\n",
                  i);
         return QZ_FAIL;
     }
@@ -859,7 +875,7 @@ int decompOutSWFallback(int i, int j, QzSession_T *sess,
                         qz_sess->next_dest,
                         &dest_receive_sz);
     if (QZ_OK != rc) {
-        QZ_ERROR("SW deCompOut fallback failure! compress fatal ERROR!\n");
+        QZ_ERROR("SW deCompOut fallback failure! decompress error!\n");
         return QZ_FAIL;
     }
 
@@ -870,4 +886,122 @@ int decompOutSWFallback(int i, int j, QzSession_T *sess,
     qz_sess->qz_out_len += dest_receive_sz;
     *dest_avail_len -= dest_receive_sz;
     return QZ_OK;
-}
\ No newline at end of file
+}
+
+/* This func is the wrapper func of compression sw fallback
+ * but with latency measure and metrixUpdate, first it would
+ * update sw latency matrix with this time latency value.
+ * Then update HW latency matrix with zero, it's for regression
+ * back to HW, in case, qatzip would always stuck on Lsm SW
+ * fallback
+ */
+int compLSMFallback(QzSession_T *sess, const unsigned char *src,
+                    unsigned int *src_len, unsigned char *dest,
+                    unsigned int *dest_len, unsigned int last)
+{
+    int rc;
+    unsigned long start_time_stamp, end_time_stamp;
+    QzSess_T *qz_sess = (QzSess_T *)sess->internal;
+
+    start_time_stamp = rdtsc();
+    /* sw fallback here */
+    rc = qzSWCompress(sess, src, src_len, dest, dest_len, last);
+
+    if (unlikely(QZ_OK != rc)) {
+        QZ_ERROR("SW LSM fallback failure! compress error!\n");
+        return QZ_FAIL;
+    }
+
+    end_time_stamp = rdtsc();
+    /* The reduction for sw latency here is for offloading more request to
+     * SW when QAT devcie is busy, maybe we could optimize this value in
+     * the future.
+     */
+    metrixUpdate(&qz_sess->SWT, (end_time_stamp - start_time_stamp) >> 2);
+    metrixUpdate(&qz_sess->RRT, 0);
+    metrixUpdate(&qz_sess->PPT, 0);
+    QZ_DEBUG("LSM mode, insert SWT %ld\n", (end_time_stamp - start_time_stamp));
+
+    return QZ_OK;
+}
+
+/* This func is the wrapper func of decompression sw fallback
+ * but with latency measure and metrixUpdate, first it would
+ * update sw latency matrix with this time latency value.
+ * Then update HW latency matrix with zero, it's for regression
+ * back to HW, in case, qatzip would always stuck on Lsm SW
+ * fallback
+ */
+int decompLSMFallback(QzSession_T *sess, const unsigned char *src,
+                      unsigned int *src_len, unsigned char *dest,
+                      unsigned int *dest_len)
+{
+    int rc;
+    unsigned long start_time_stamp, end_time_stamp;
+    QzSess_T *qz_sess = (QzSess_T *)sess->internal;
+
+    start_time_stamp = rdtsc();
+    /* sw fallback here */
+    rc = qzSWDecompressMulti(sess, src, src_len, dest, dest_len);
+
+    if (unlikely(QZ_OK != rc)) {
+        QZ_ERROR("SW LSM fallback failure! decompress error!\n");
+        return QZ_FAIL;
+    }
+
+    end_time_stamp = rdtsc();
+    /* The reduction for sw latency here is for offloading more request to
+     * SW when QAT devcie is busy, maybe we could optimize this value in
+     * the future.
+     */
+    metrixUpdate(&qz_sess->SWT, (end_time_stamp - start_time_stamp) >> 2);
+    metrixUpdate(&qz_sess->RRT, 0);
+    metrixUpdate(&qz_sess->PPT, 0);
+    QZ_DEBUG("LSM mode, insert SWT %ld\n", (end_time_stamp - start_time_stamp));
+    return QZ_OK;
+}
+
+int AsyncDecompOutSWFallback(int i, int j, QzSession_T *sess,
+                             QzAsyncReq_T *req)
+{
+    int rc;
+    QzSess_T *qz_sess = (QzSess_T *)sess->internal;
+    DataFormatInternal_T data_fmt = qz_sess->sess_params.data_fmt;
+    unsigned int src_send_sz =
+        g_process.qz_inst[i].src_buffers[j]->pBuffers->dataLenInBytes;
+    unsigned int dest_receive_sz =
+        g_process.qz_inst[i].dest_buffers[j]->pBuffers->dataLenInBytes;
+    /*  pinned buffer ptr already strip header, need retrieve.
+    *   use original src_ptr and consumed len to get src_ptr for this request.
+    */
+    const unsigned char *src_ptr = req->src + req->req_in_len;
+    src_send_sz += (outputHeaderSz(data_fmt) + outputFooterSz(data_fmt)) ;
+
+    if (!qz_sess->sess_params.sw_backup) {
+        QZ_DEBUG("The instance %d heartbeat is failure, Don't enable sw fallback, decompressOut fatal ERROR!\n",
+                 i);
+        return QZ_FAIL;
+    }
+
+    QZ_DEBUG("The request get dummy empty respond, offload to software!\n");
+    QZ_DEBUG("SW DecompOut src_ptr %p, dest_ptr %p, Sending %u bytes, receive %u bytes, seq = %ld\n",
+             src_ptr, req->dest, src_send_sz, dest_receive_sz,
+             g_process.qz_inst[i].stream[j].seq);
+    rc = qzSWDecompress(sess,
+                        src_ptr,
+                        &src_send_sz,
+                        req->dest,
+                        &dest_receive_sz);
+    if (QZ_OK != rc) {
+        QZ_ERROR("SW deCompOut fallback failure! compress fatal ERROR!\n");
+        return QZ_FAIL;
+    }
+
+    /* update qz_sess info, Don't need clean dest buffer */
+    decompOutErrorDestBufferCleanUp(i, j);
+    req->dest += dest_receive_sz;
+    req->req_in_len += src_send_sz;
+    req->req_out_len += dest_receive_sz;
+    req->qzResults->dest_len -= dest_receive_sz;
+    return QZ_OK;
+}
