@@ -145,16 +145,15 @@ static void doDumpThreadInfo(ThreadList_T *node,
             serv_title = "Decompression";
         }
 
-        QZ_PRINT("[INFO]: %s num_th %u\n",
-                 serv_title, num_node);
+        QZ_INFO("%s num_th %u\n", serv_title, num_node);
         while (node) {
-            QZ_PRINT("th_id: %u comp_hw_count: %u comp_sw_count: %u "
-                     "decomp_hw_count: %u decomp_sw_count: %u\n",
-                     node->thread_id,
-                     node->comp_hw_count,
-                     node->comp_sw_count,
-                     node->decomp_hw_count,
-                     node->decomp_sw_count);
+            QZ_INFO("th_id: %u comp_hw_count: %u comp_sw_count: %u "
+                    "decomp_hw_count: %u decomp_sw_count: %u\n",
+                    node->thread_id,
+                    node->comp_hw_count,
+                    node->comp_sw_count,
+                    node->decomp_hw_count,
+                    node->decomp_sw_count);
             i++;
             node = node->next;
             if (i == num_node) {
@@ -165,7 +164,7 @@ static void doDumpThreadInfo(ThreadList_T *node,
         if (node) {
             QZ_ERROR("[ERROR]: there's node left in the list\n");
         }
-        QZ_PRINT("\n");
+        QZ_INFO("\n");
     }
 }
 
@@ -181,12 +180,87 @@ void dumpThreadInfo(void)
 }
 #endif
 
+QzLogLevel_T currentLogLevel = LOG_WARNING;
+
+QzLogLevel_T qzSetLogLevel(QzLogLevel_T level)
+{
+    QzLogLevel_T oldLogLevel = currentLogLevel;
+    currentLogLevel = level;
+    return oldLogLevel;
+}
+
+const char *getLogLevelString(QzLogLevel_T level)
+{
+    switch (level) {
+    case LOG_ERROR:
+        return "Error";
+    case LOG_WARNING:
+        return "Warning";
+    case LOG_INFO:
+        return "Info";
+    case LOG_DEBUG1:
+        return "Debug";
+    case LOG_DEBUG2:
+        return "Test";
+    case LOG_DEBUG3:
+        return "Memory";
+    default:
+        return "UNKNOWN";
+    }
+}
+
+void logMessage(QzLogLevel_T level, const char *file, int line,
+                const char *format, ...)
+{
+
+    if (level > currentLogLevel) return;
+
+    time_t now = time(NULL);
+    struct tm *tm_info = localtime(&now);
+    char time_buffer[26];
+    strftime(time_buffer, 26, "%Y-%m-%d %H:%M:%S", tm_info);
+    FILE *output = stderr;
+
+    switch (level) {
+    case LOG_NONE:
+        output = stdout;
+        break;
+    case LOG_ERROR:
+    case LOG_WARNING:
+        output = stderr;
+        fprintf(output, "[%s] [%s] (%s:%d): ", getLogLevelString(level), time_buffer,
+                file, line);
+        break;
+    case LOG_INFO:
+    case LOG_DEBUG1:
+    case LOG_DEBUG2:
+    case LOG_DEBUG3:
+        output = stdout;
+        fprintf(output, "[%s]: ", getLogLevelString(level));
+        break;
+    default:
+        break;
+    }
+
+    va_list args;
+    va_start(args, format);
+    vfprintf(output, format, args);
+    va_end(args);
+}
+
 void initDebugLock(void)
 {
     pthread_mutex_init(&g_qat_thread.comp_lock, NULL);
     pthread_mutex_init(&g_qat_thread.decomp_lock, NULL);
 }
 
+/* LSM TODO:
+ * expose global variabled (mention variable name) to API Level to tune
+ * LSM Mode Performance lsm_sw_met_seed is the emprical mean of sw Latence
+ */
+unsigned int lsm_met_len_shift = 6;
+unsigned int lsm_sw_met_seed = 1000;
+
 static int qzSetupDcSessionData(CpaDcSessionSetupData *session_setup_data,
                                 QzSessionParamsInternal_T *params)
 {
@@ -200,8 +274,13 @@ static int qzSetupDcSessionData(CpaDcSessionSetupData *session_setup_data,
     case DEFLATE_GZIP:
     case DEFLATE_GZIP_EXT:
     case DEFLATE_RAW:
+    case DEFLATE_ZLIB:
         session_setup_data->compType = CPA_DC_DEFLATE;
-        session_setup_data->checksum = CPA_DC_CRC32;
+        if (params->data_fmt == DEFLATE_ZLIB) {
+            session_setup_data->checksum = CPA_DC_ADLER32;
+        } else {
+            session_setup_data->checksum = CPA_DC_CRC32;
+        }
         session_setup_data->autoSelectBestHuffmanTree =
             CPA_DC_ASB_UNCOMP_STATIC_DYNAMIC_WITH_STORED_HDRS;
         if (params->huffman_hdr == QZ_DYNAMIC_HDR) {
@@ -277,6 +356,14 @@ int qzSetupSessionInternal(QzSession_T *sess)
         return rc;
     }
 
+    /* LSM setup */
+    if (qz_sess->sess_params.is_sensitive_mode == true) {
+        metrixReset(&qz_sess->RRT);
+        metrixReset(&qz_sess->PPT);
+        metrixReset(&qz_sess->SWT);
+        qz_sess->SWT.arr_avg = lsm_sw_met_seed;
+    }
+
     qz_sess->inst_hint = -1;
     qz_sess->seq = 0;
     qz_sess->seq_in = 0;
@@ -451,6 +538,47 @@ int qzCheckParamsDeflate(QzSessionParamsDeflate_T *params)
     return QZ_OK;
 }
 
+int qzCheckParamsDeflateExt(QzSessionParamsDeflateExt_T *params)
+{
+    assert(params);
+    if (qzCheckParamsCommon(&params->deflate_params.common_params) != QZ_OK) {
+        return QZ_PARAMS;
+    }
+
+    if (params->deflate_params.common_params.comp_algorithm != QZ_DEFLATE) {
+        QZ_ERROR("Invalid comp_algorithm value\n");
+        return QZ_PARAMS;
+    }
+
+    if (params->deflate_params.huffman_hdr > QZ_STATIC_HDR) {
+        QZ_ERROR("Invalid huffman_hdr value\n");
+        return QZ_PARAMS;
+    }
+
+    if ((params->deflate_params.common_params.comp_lvl < QZ_LZS_COMP_LVL_MINIMUM) ||
+        (params->deflate_params.common_params.comp_lvl >
+         QZ_DEFLATE_COMP_LVL_MAXIMUM_Gen3)) {
+        QZ_ERROR("Invalid comp_lvl value\n");
+        return QZ_PARAMS;
+    }
+    QZ_DEBUG(" data_fmt value = %d , zlib_format = %d ",
+             params->deflate_params.data_fmt, params->zlib_format);
+    if ((params->zlib_format == 0) &&
+        (params->deflate_params.data_fmt > QZ_DEFLATE_RAW)) {
+        QZ_ERROR("Invalid data_fmt value\n");
+        return QZ_PARAMS;
+    }
+
+    return QZ_OK;
+    /*
+        if (qzCheckParamsDeflate(&params->deflate_params) != QZ_OK) {
+            return QZ_PARAMS;
+        }
+
+        return QZ_OK;
+    */
+}
+
 int qzCheckParamsLZ4(QzSessionParamsLZ4_T *params)
 {
     assert(params);
@@ -583,6 +711,25 @@ void qzSetParamsDeflate(QzSessionParamsDeflate_T *params,
     internal_params->huffman_hdr = params->huffman_hdr;
 }
 
+/**
+ * Set deflate ext session params into internal params
+ */
+void qzSetParamsDeflateExt(QzSessionParamsDeflateExt_T *params,
+                           QzSessionParamsInternal_T *internal_params)
+{
+    assert(params);
+    assert(internal_params);
+
+    qzSetParamsDeflate(&params->deflate_params, internal_params);
+
+    if (params->zlib_format == 1) {
+        internal_params->data_fmt = DEFLATE_ZLIB;
+        internal_params->zlib_format = 1;
+    }
+    internal_params->stop_decompression_stream_end =
+        params->stop_decompression_stream_end;
+}
+
 void qzSetParamsLZ4(QzSessionParamsLZ4_T *params,
                     QzSessionParamsInternal_T *internal_params)
 {
@@ -687,6 +834,25 @@ void qzGetParamsDeflate(QzSessionParamsInternal_T *internal_params,
     params->common_params.comp_algorithm = QZ_DEFLATE;
 }
 
+/**
+ * Get deflateExt session params from internal params
+ */
+void qzGetParamsDeflateExt(QzSessionParamsInternal_T *internal_params,
+                           QzSessionParamsDeflateExt_T *params)
+{
+    assert(params);
+    assert(internal_params);
+
+    qzGetParamsDeflate(internal_params, &params->deflate_params);
+    if (internal_params->data_fmt == DEFLATE_ZLIB) {
+        internal_params->zlib_format = 1;
+    }
+
+    params->stop_decompression_stream_end =
+        internal_params->stop_decompression_stream_end;
+    params->zlib_format = internal_params->zlib_format;
+}
+
 /**
  * Get LZ4 session params from internal params
  */
@@ -734,6 +900,9 @@ inline unsigned long outputFooterSz(DataFormatInternal_T data_fmt)
     case LZ4S_BK:
         size = 0;
         break;
+    case DEFLATE_ZLIB:
+        size = stdZlibFooterSz();
+        break;
     case DEFLATE_GZIP_EXT:
     default:
         size = stdGzipFooterSz();
@@ -762,6 +931,9 @@ unsigned long outputHeaderSz(DataFormatInternal_T data_fmt)
     case LZ4S_BK:
         size = qzLZ4SBlockHeaderSz();
         break;
+    case DEFLATE_ZLIB:
+        size = stdZlibHeaderSz();
+        break;
     case DEFLATE_GZIP_EXT:
     default:
         size = qzGzipHeaderSz();
@@ -775,8 +947,6 @@ void outputHeaderGen(unsigned char *ptr,
                      CpaDcRqResults *res,
                      DataFormatInternal_T data_fmt)
 {
-    QZ_DEBUG("Generate header\n");
-
     switch (data_fmt) {
     case DEFLATE_4B:
         qz4BHeaderGen(ptr, res);
@@ -792,6 +962,9 @@ void outputHeaderGen(unsigned char *ptr,
     case LZ4S_BK:
         qzLZ4SBlockHeaderGen(ptr, res);
         break;
+    case DEFLATE_ZLIB:
+        stdZlibHeaderGen(ptr, res);
+        break;
     case DEFLATE_GZIP_EXT:
     default:
         qzGzipHeaderGen(ptr, res);
@@ -799,11 +972,10 @@ void outputHeaderGen(unsigned char *ptr,
     }
 }
 
-inline void outputFooterGen(QzSess_T *qz_sess,
+inline void outputFooterGen(unsigned char *ptr,
                             CpaDcRqResults *res,
                             DataFormatInternal_T data_fmt)
 {
-    unsigned char *ptr = qz_sess->next_dest;
     switch (data_fmt) {
     case DEFLATE_RAW:
         break;
@@ -812,6 +984,9 @@ inline void outputFooterGen(QzSess_T *qz_sess,
         break;
     case LZ4S_BK:
         break;
+    case DEFLATE_ZLIB:
+        qzZlibFooterGen(ptr, res);
+        break;
     case DEFLATE_GZIP_EXT:
     default:
         qzGzipFooterGen(ptr, res);
@@ -835,11 +1010,20 @@ int isQATProcessable(const unsigned char *ptr,
     case DEFLATE_4B:
     case DEFLATE_GZIP:
     case DEFLATE_GZIP_EXT:
+    case DEFLATE_ZLIB:
         rc = isQATDeflateProcessable(ptr, src_len, qz_sess);
         break;
     case LZ4_FH:
         rc = isQATLZ4Processable(ptr, src_len, qz_sess);
         break;
+    case DEFLATE_RAW:
+        if (*src_len < qz_sess->sess_params.input_sz_thrshold) {
+            QZ_DEBUG("isQATProcessable: deflate_raw src_len is less than input threshhold\n");
+            rc = 0;
+        } else {
+            rc = 1;
+        }
+        break;
     default:
         rc = 0;
         break;
@@ -908,7 +1092,8 @@ void compBufferSetup(int i, int j, QzSess_T *qz_sess,
     dest_receive_sz = *qz_sess->dest_sz > DEST_SZ(src_send_sz) ?
                       DEST_SZ(src_send_sz) - outputHeaderSz(data_fmt) :
                       *qz_sess->dest_sz - outputHeaderSz(data_fmt);
-
+    QZ_DEBUG("compBufferSetup dest_receive_sz = %d , src_send_sz = %d\n",
+             dest_receive_sz, src_send_sz);
     g_process.qz_inst[i].src_buffers[j]->pBuffers->dataLenInBytes = src_send_sz;
     g_process.qz_inst[i].dest_buffers[j]->pBuffers->dataLenInBytes =
         dest_receive_sz;
@@ -1082,6 +1267,8 @@ int checkHeader(QzSess_T *qz_sess, unsigned char *src,
         if ((unsigned char *)qzFooter == src_ptr + src_avail_len - stdGzipFooterSz()) {
             isEndWithFooter = 1;
         }
+        QZ_DEBUG("checkHeader: DEFLATE_GZIP compressed_sz = %u , uncompressed_sz = %u \n",
+                 compressed_sz, uncompressed_sz);
         break;
     case DEFLATE_GZIP_EXT:
         if (QZ_OK != qzGzipHeaderExt(src_ptr, hdr)) {
@@ -1089,6 +1276,9 @@ int checkHeader(QzSess_T *qz_sess, unsigned char *src,
         }
         compressed_sz = (long)(hdr->extra.qz_e.dest_sz);
         uncompressed_sz = (long)(hdr->extra.qz_e.src_sz);
+        QZ_DEBUG("checkHeader: DEFLATE_GZIP_EXT compressed_sz = %u , uncompressed_sz = %u \n",
+                 compressed_sz, uncompressed_sz);
+
         break;
     case DEFLATE_4B:
         compressed_sz = *(int *)src_ptr;
@@ -1110,6 +1300,19 @@ int checkHeader(QzSess_T *qz_sess, unsigned char *src,
         compressed_sz = (unsigned char *)lz4Footer - src_ptr - qzLZ4HeaderSz();
         uncompressed_sz = qzLZ4Header->cnt_size;
         break;
+    case DEFLATE_RAW:
+        compressed_sz = *(qz_sess->src_sz);
+        uncompressed_sz = *(qz_sess->dest_sz);
+        QZ_DEBUG("checkHeader: DEFLATE_RAW HW Decompression enabled uncompressed_sz is %u \n",
+                 uncompressed_sz);
+        break;
+    case DEFLATE_ZLIB:
+        compressed_sz = src_avail_len - stdZlibFooterSz() - stdZlibHeaderSz();
+        uncompressed_sz = (qz_sess->sess_params.hw_buff_sz > dest_avail_len) ?
+                          dest_avail_len : qz_sess->sess_params.hw_buff_sz;
+        QZ_DEBUG("checkHeader DEFLATE_ZLIB: compressed_sz %u  uncompressed_sz %u \n",
+                 compressed_sz, uncompressed_sz);
+        break;
     default:
         return QZ_FAIL;
     }
@@ -1154,6 +1357,7 @@ void decompBufferSetup(int i, int j, QzSess_T *qz_sess,
 {
     StdGzF_T *qzFooter = NULL;
     QzLZ4F_T *lz4Footer = NULL;
+    StdZlibF_T *zlibFooter = NULL;
     unsigned int src_send_sz = 0;
     unsigned int dest_receive_sz = 0;
     int src_mem_type = qzMemFindAddr(src_ptr);
@@ -1179,7 +1383,7 @@ void decompBufferSetup(int i, int j, QzSess_T *qz_sess,
 
     QZ_DEBUG("doDecompressIn: Sending %u bytes starting at 0x%lx\n",
              src_send_sz, (unsigned long)src_ptr);
-    QZ_DEBUG("sending seq number %lu %d %ld\n", i, j, qz_sess->seq);
+    QZ_DEBUG("sending seq number %d %d %ld\n", i, j, qz_sess->seq);
 
     if (DEFLATE_GZIP_EXT == data_fmt ||
         DEFLATE_GZIP == data_fmt) {
@@ -1190,6 +1394,11 @@ void decompBufferSetup(int i, int j, QzSess_T *qz_sess,
         lz4Footer = (QzLZ4F_T *)(src_ptr + src_send_sz); //TODO
         g_process.qz_inst[i].stream[j].checksum = lz4Footer->cnt_cksum;
         g_process.qz_inst[i].stream[j].orgdatalen = hdr->extra.qz_e.src_sz;
+    } else if (DEFLATE_ZLIB == data_fmt) {
+        zlibFooter = (StdZlibF_T *)(src_ptr + src_send_sz);
+        g_process.qz_inst[i].stream[j].checksum = zlibFooter->adler32;
+        g_process.qz_inst[i].stream[j].orgdatalen =
+            hdr->extra.qz_e.src_sz; //uncompressed size stored in checkheader
     }
 
     /*set up src dest buffers*/
@@ -1275,14 +1484,41 @@ int decompOutCheckSum(int i, int j, QzSession_T *sess,
                       CpaDcRqResults *resl)
 {
     QzSess_T *qz_sess = (QzSess_T *)sess->internal;
+    //alder32_swap to read the checksum in proper order.
+    if (qz_sess->sess_params.data_fmt == DEFLATE_ZLIB) {
+        if (resl->consumed <
+            g_process.qz_inst[i].src_buffers[j]->pBuffers->dataLenInBytes) {
+            // we need to refetch trailer from the end of consumed length which is the stream end
+            unsigned char *src_ptr = qz_sess->src + qz_sess->qz_in_len + outputHeaderSz(
+                                         qz_sess->sess_params.data_fmt);
+            StdZlibF_T *zlibFooter = (StdZlibF_T *)(src_ptr + resl->consumed);
+            g_process.qz_inst[i].stream[j].checksum = zlibFooter->adler32;
+            if (qz_sess->single_thread &&
+                (qz_sess->sess_params.stop_decompression_stream_end == 0)) {
+                qz_sess->last_submitted = 0;
+            }
+            QZ_DEBUG("src_ptr  %p  qz_sess->qz_in_len %lu data checksum %x\n",
+                     src_ptr, qz_sess->qz_in_len,
+                     g_process.qz_inst[i].stream[j].checksum);
+        } else {
+            qz_sess->last_submitted = 1;
+        }
+        g_process.qz_inst[i].stream[j].checksum = ntohl(
+                    g_process.qz_inst[i].stream[j].checksum);
+        QZ_DEBUG("After resp checksum: %x data checksum %x\n", resl->checksum,
+                 g_process.qz_inst[i].stream[j].checksum);
+    }
     if ((qz_sess->sess_params.data_fmt != DEFLATE_4B) &&
+        (qz_sess->sess_params.data_fmt != DEFLATE_RAW) &&
         unlikely(resl->checksum !=
                  g_process.qz_inst[i].stream[j].checksum ||
-                 resl->produced != g_process.qz_inst[i].stream[j].orgdatalen)) {
+                 (resl->produced != g_process.qz_inst[i].stream[j].orgdatalen &&
+                  qz_sess->sess_params.data_fmt != DEFLATE_ZLIB))) {
         QZ_ERROR("Error in check footer, inst %d, stream %d\n", i, j);
-        QZ_DEBUG("resp checksum: %x data checksum %x\n",
+        QZ_DEBUG("resp checksum: %x data checksum %x data_fmt %d\n",
                  resl->checksum,
-                 g_process.qz_inst[i].stream[j].checksum);
+                 g_process.qz_inst[i].stream[j].checksum,
+                 qz_sess->sess_params.data_fmt);
         QZ_DEBUG("resp produced :%d data produced: %d\n",
                  resl->produced,
                  g_process.qz_inst[i].stream[j].orgdatalen);
@@ -1294,3 +1530,294 @@ int decompOutCheckSum(int i, int j, QzSession_T *sess,
     }
     return QZ_OK;
 }
+
+void setDeflateEndOfStream(QzSess_T *qz_sess, unsigned char val)
+{
+    QzDeflateExtCustomData_T  *data = (QzDeflateExtCustomData_T *)
+                                      qz_sess->qzdeflateExtData;
+    if (data != NULL) {
+        data->end_of_stream = val;
+        QZ_DEBUG("\tHW setDeflateEndOfStream: setting end_of_stream to %d \n", val);
+    }
+}
+
+unsigned char getDeflateEndOfStream(QzSess_T *qz_sess)
+{
+    QzDeflateExtCustomData_T  *data = (QzDeflateExtCustomData_T *)
+                                      qz_sess->qzdeflateExtData;
+    if (data != NULL) {
+        QZ_DEBUG("\tHW getDeflateEndOfStream:  end_of_stream  %d \n",
+                 data->end_of_stream);
+        return data->end_of_stream;
+    }
+    return 0;
+}
+
+inline void metrixReset(LatencyMetrix_T *m)
+{
+    if (m == NULL) {
+        return;
+    }
+    if (m->latency_array == NULL) {
+        m->latency_array = calloc(LSM_MET_DEPTH, sizeof(*(m->latency_array)));
+    }
+
+    for (int i = 0; i < LSM_MET_DEPTH; i++) {
+        m->latency_array[i] = 0;
+    }
+    m->arr_total = 0;
+    m->arr_avg = 0;
+    m->arr_idx = 0;
+#ifdef QATZIP_DEBUG
+    m->invoke_counter = 0;
+    m->sess_lat_total = 0;
+    m->sess_lat_avg = 0;
+#endif
+}
+
+inline void metrixUpdate(LatencyMetrix_T *m, unsigned long val)
+{
+    if (m == NULL) {
+        return;
+    }
+
+    m->arr_total -= m->latency_array[m->arr_idx];
+    m->latency_array[m->arr_idx] = val;
+    m->arr_total += m->latency_array[m->arr_idx];
+    m->arr_avg = m->arr_total >> lsm_met_len_shift;
+
+    if (++m->arr_idx >= LSM_MET_DEPTH) {
+        m->arr_idx = 0;
+    }
+
+    if (val != 0) {
+        m->invoke_counter++;
+    }
+#ifdef QATZIP_DEBUG
+    m->sess_lat_total += val;
+
+    if (m->invoke_counter != 0) {
+        m->sess_lat_avg = (m->sess_lat_total) / (m->invoke_counter);
+    }
+
+    if (val < m->min_latency) {
+        m->min_latency = val;
+        m->min_cnt = m->invoke_counter;
+    }
+    if (val > m->min_latency) {
+        m->max_latency = val;
+        m->max_cnt = m->invoke_counter;
+    }
+#endif
+
+    QZ_INFO("The latency for request %lu is %lu, avg is %lu\n", m->arr_idx, val,
+            m->arr_avg);
+    return;
+}
+
+int AsyncCompOutCheckDestLen(int i, int j, QzSession_T *sess,
+                             long dest_receive_sz)
+{
+    QzSess_T *qz_sess = (QzSess_T *)sess->internal;
+    QzAsyncReq_T *req = g_process.qz_inst[i].stream[j].req;
+    long remaining = req->qzResults->dest_len - req->req_out_len - dest_receive_sz;
+    if (unlikely(remaining < 0)) {
+        compOutSkipErrorRespond(i, j, qz_sess);
+        // qz_sess->stop_submitting = 1;
+        sess->thd_sess_stat = QZ_BUF_ERROR;
+        return sess->thd_sess_stat;
+    }
+    return QZ_OK;
+}
+
+void AsyncCompOutValidDestBufferCleanUp(int i, int j,
+                                        unsigned int dest_receive_sz)
+{
+    QzAsyncReq_T *req = g_process.qz_inst[i].stream[j].req;
+    CpaBoolean need_cont_mem =
+        g_process.qz_inst[i].instance_info.requiresPhysicallyContiguousMemory;
+
+    if (!need_cont_mem) {
+        QZ_DEBUG("Compress SVM Enabled in doCompressOut\n");
+    }
+
+    if (g_process.qz_inst[i].stream[j].dest_need_reset) {
+        g_process.qz_inst[i].dest_buffers[j]->pBuffers->pData =
+            g_process.qz_inst[i].stream[j].orig_dest;
+        g_process.qz_inst[i].stream[j].dest_need_reset = 0;
+    } else {
+        QZ_MEMCPY(req->dest,
+                  g_process.qz_inst[i].dest_buffers[j]->pBuffers->pData,
+                  req->qzResults->dest_len - req->req_out_len,
+                  dest_receive_sz);
+    }
+}
+
+void AsyncDecompOutValidDestBufferCleanUp(int i, int j, QzSess_T *qz_sess,
+        CpaDcRqResults *resl,
+        QzAsyncReq_T *req)
+{
+    if (!g_process.qz_inst[i].stream[j].dest_need_reset) {
+        QZ_DEBUG("memory copy in doDecompressOut\n");
+        QZ_MEMCPY(req->dest,
+                  g_process.qz_inst[i].dest_buffers[j]->pBuffers->pData,
+                  req->qzResults->dest_len,
+                  resl->produced);
+    } else {
+        g_process.qz_inst[i].dest_buffers[j]->pBuffers->pData =
+            g_process.qz_inst[i].stream[j].orig_dest;
+        g_process.qz_inst[i].stream[j].dest_need_reset = 0;
+    }
+}
+
+static int QzRingMoveProdHead(QzRing_T *ring, uint32_t *old_head,
+                              uint32_t *new_head, int is_single_producer)
+{
+    const uint32_t capacity = ring->capacity;
+    int success = 0;
+    uint32_t free_entries = 0;
+
+    do {
+        *old_head = ring->prod.head;
+        *new_head = *old_head + 1;
+
+        free_entries = (capacity + ring->cons.tail - *old_head);
+        if (1 > free_entries)
+            return QZ_FAIL;
+
+        if (is_single_producer) {
+            ring->prod.head = *new_head;
+            success = 1;
+        } else {
+            success = __sync_bool_compare_and_swap((uint32_t *)(uintptr_t)&ring->prod.head,
+                                                   *old_head, *new_head);
+        }
+    } while (success == 0);
+    return QZ_OK;
+}
+
+static void QzRingEnqueueElem(QzRing_T *ring, uint32_t prod_head, void *obj)
+{
+    prod_head = prod_head % ring->capacity;
+    ring->elems[prod_head] = obj;
+}
+
+static void QzRingUpdatTail(QzRingHeadTail_T *ht, uint32_t old_val,
+                            uint32_t new_val, uint32_t single)
+{
+    if (!single)
+        __sync_bool_compare_and_swap((volatile uint32_t *)(uintptr_t)&ht->tail, old_val,
+                                     new_val);
+    else
+        ht->tail = new_val;
+}
+
+static int QzRingMoveConsHead(QzRing_T *ring,
+                              uint32_t *old_head, uint32_t *new_head, int is_single_consumer)
+{
+    int success;
+    int entries = 0;
+
+    do {
+        *old_head = ring->cons.head;
+
+        entries = (ring->prod.tail - *old_head);
+
+        if (1 > entries)
+            return QZ_FAIL;
+
+        *new_head = *old_head + 1;
+        if (is_single_consumer) {
+            ring->cons.head = *new_head;
+            success = 1;
+        } else {
+            success = __sync_bool_compare_and_swap((uint32_t *)(uintptr_t)&ring->cons.head,
+                                                   *old_head, *new_head);
+        }
+    } while (success == 0);
+    return QZ_OK;
+}
+
+static void *QzRingDequeueElem(QzRing_T *ring, uint32_t cons_head)
+{
+    void *obj;
+    cons_head = cons_head % ring->capacity;
+    obj = ring->elems[cons_head];
+    ring->elems[cons_head] = NULL;
+    return obj;
+}
+
+QzRing_T *QzRingCreate(int size)
+{
+    if (size < 1) {
+        QZ_ERROR("Create ring size is incorrect\n");
+        return NULL;
+    }
+    QzRing_T *ring;
+    ring = (QzRing_T *)calloc(1, sizeof(QzRing_T));
+    ring->elems = (void *)calloc(size, sizeof(void *));
+    ring->size = size;
+    ring->mask = size - 1;
+    ring->capacity = size;
+    return ring;
+}
+
+void QzClearRing(QzRing_T *ring)
+{
+    for (int i = 0; i < ring->size; i++) {
+        if (NULL != ring->elems[i]) {
+            free(ring->elems[i]);
+            ring->elems[i] = NULL;
+        }
+    }
+    ring->mask = ring->size - 1;
+    ring->capacity = ring->size;
+    ring->cons.head = 0;
+    ring->cons.tail = 0;
+    ring->prod.head = 0;
+    ring->prod.tail = 0;
+}
+
+void QzRingFree(QzRing_T *ring)
+{
+    if (ring != NULL) {
+        if (ring->elems != NULL) {
+            free(ring->elems);
+        }
+        free(ring);
+    }
+}
+
+int QzRingProduceEnQueue(QzRing_T *ring, void *obj, int is_single_producer)
+{
+    uint32_t prod_head, prod_next;
+    int rc = QZ_OK;
+
+    rc = QzRingMoveProdHead(ring, &prod_head, &prod_next, is_single_producer);
+    if (QZ_FAIL == rc)
+        return rc;
+
+    QzRingEnqueueElem(ring, prod_head, obj);
+
+    QzRingUpdatTail(&ring->prod, prod_head, prod_next, is_single_producer);
+    return rc;
+}
+
+void *QzRingConsumeDequeue(QzRing_T *ring, int is_single_consumer)
+{
+    uint32_t cons_head = 0;
+    uint32_t cons_next = 0;
+    // uint32_t entries;
+    int rc = 0;
+    void *obj = NULL;
+
+    rc = QzRingMoveConsHead(ring, &cons_head, &cons_next, is_single_consumer);
+    if (QZ_FAIL == rc)
+        return NULL;
+
+    obj = QzRingDequeueElem(ring, cons_head);
+
+    QzRingUpdatTail(&ring->cons, cons_head, cons_next, is_single_consumer);
+
+    return obj;
+}
