@@ -46,7 +46,7 @@ QzSessionParamsLZ4S_T sess_params_zstd_default = {
     .common_params.hw_buff_sz        = QZ_HW_BUFF_SZ,
     .common_params.strm_buff_sz      = QZ_STRM_BUFF_SZ_DEFAULT,
     .common_params.input_sz_thrshold = QZ_COMP_THRESHOLD_DEFAULT,
-    .common_params.req_cnt_thrshold  = 32,
+    .common_params.req_cnt_thrshold  = QZ_REQ_THRESHOLD_DEFAULT,
     .common_params.wait_cnt_thrshold = QZ_WAIT_CNT_THRESHOLD_DEFAULT,
     .common_params.polling_mode   = QZ_PERIODICAL_POLLING,
     .lz4s_mini_match   = 3,
@@ -64,6 +64,7 @@ void qzstd_help()
         "  -d,       decompress",
         "  -h,       show help information",
         "  -L,       set compression level of QAT",
+        "  -g,       set qatzip loglevel(none|error|warn|info|debug)",
         "  -o,       set output file name",
         "  -C,       set zstd block size && QAT hw buffer size",
         "  -r,       set max in-flight request number",
@@ -367,7 +368,7 @@ int compressFile(int in_file, int out_file)
 
     do {
         bytes_read = read(in_file, src_buffer, src_buffer_size);
-        QZ_PRINT("Reading input file (%u Bytes)\n", bytes_read);
+        QZ_INFO("Reading input file (%u Bytes)\n", bytes_read);
         puts("compressing...");
 
         dst_buffer_size = max_dst_buffer_size;
@@ -474,7 +475,7 @@ int decompressFile(int in_file, int out_file)
 
     do {
         bytes_read = read(in_file, src_buffer, src_buffer_size);
-        QZ_PRINT("Reading input file (%lu Bytes)\n", bytes_read);
+        QZ_INFO("Reading input file (%lu Bytes)\n", bytes_read);
         puts("Decompressing...");
 
         ZSTD_inBuffer input = { src_buffer, bytes_read, 0 };
